/* vb_dbg.dll Debugger controller written for VB
   from code @ http://sandsprite.com

   Portions of this code were taken from the DEB debugger
   example in the platform SDK. HWBP routines from a newsgroup
   but dont seem to be working just yet.
*/

#include <windows.h>
#include <string.h>
#include <stdio.h>
#include <StdDef.H> // offsetof()

#define DBG_COND_EXECUTE 0 
#define DBG_COND_WRITE  1 
#define DBG_COND_READWRITE  3 

int  targetHWND = 0;
int  pid = 0;
int  isDebugging = 0 ;
int  debugDecision = 0 ;

char proc[255];
char errMsg[500];

DEBUG_EVENT  DebugEvent;
HANDLE		 hDebugEventThread;


typedef struct{
    int dwFlag;
    int cbSize;
    int lpData;
} cpyData;



DWORD WINAPI DebugEventThread(void*);

void __stdcall SetTargetHwnd(int hwnd){ targetHWND = hwnd;}
void __stdcall GetErrorMessage(char* buf500){ strcpy(buf500,errMsg);}
void __stdcall Continue(){ 	debugDecision = DBG_CONTINUE; }
void __stdcall StopDebugger(){ 	TerminateThread(hDebugEventThread,0); isDebugging = 0;}
int  __stdcall ActivePID(){ if(isDebugging==1){return pid;}else{return 0;}}

inline void SETBITS(unsigned long& dw, int lowBit, int bits, int newValue){ 
	int mask = (1 << bits) - 1; 
	dw = (dw & ~(mask << lowBit)) | (newValue << lowBit); 
} 

// Address   - address to break on. 
// Length    - must be 0, 1, 2 or 4 
//  Condition - must be 0, 1 or 3  (i.e. DBG_COND_xxx) 
// 
//  returns -1 (error), or a zero-based index which 
//  specifies which debug register was used....use this 
//  value in your call to ClearHardwareBP 
// 
int __stdcall SetHardwareBP(HANDLE hThread, int Address, int Condition){ 
	
	 int i, Length=1;
	 CONTEXT context = { CONTEXT_DEBUG_REGISTERS }; 
	 
	 // get contents of every debug register 
	 if(!GetThreadContext(hThread, &context)) return -1; 

	 // find available hardware register 
	 for(i = 0; i < 4; i++){ 
		  if((context.Dr7 & (1 << (i * 2))) == 0){ 
			   *(&context.Dr0 + i) = Address; 
			   SETBITS(context.Dr7, 16 + i*4, 2, Condition); 
			   SETBITS(context.Dr7, 18 + i*4, 2, Length); 
			   SETBITS(context.Dr7, i * 2,    1, 1); 
			   if(!SetThreadContext(hThread, &context)) return -1; 
			   return i; 
		  } 
     } 
	 return -1;
}

BOOL __stdcall ClearHardwareBP(HANDLE hThread, int DbgRegister){
	 CONTEXT context = { CONTEXT_DEBUG_REGISTERS }; 
	 if(!GetThreadContext(hThread, &context)) return FALSE; 
	 SETBITS(context.Dr7, DbgRegister*2, 1, 0); 
	 if(!SetThreadContext(hThread, &context)) return FALSE; 
	 return TRUE; 
} 


int DbgMsg(DEBUG_EVENT *Buffer){
  
  if(targetHWND==0){
	  MessageBox(0,"SetTargetHWND First!","",0);
	  return 0;
  }
  
  cpyData cpStructData;
  
  cpStructData.cbSize = sizeof(DEBUG_EVENT);
  cpStructData.lpData = (int)Buffer;
  cpStructData.dwFlag = 3;
  
  return SendMessage((HWND)targetHWND, WM_COPYDATA, (WPARAM)targetHWND,(LPARAM)&cpStructData);

} 


int __stdcall LaunchDebugedProcess(char* filename){
	
   if(isDebugging==1){
		sprintf(errMsg,"Only one debuggee per session!");
		return 0;
   }
			
   pid=0;	
   strcpy(proc,filename);
 
   if( !( hDebugEventThread = CreateThread(
 							   (LPSECURITY_ATTRIBUTES) 0L,
							   (DWORD) 0,
							   (LPTHREAD_START_ROUTINE) DebugEventThread,
							   (LPVOID) 0,
							   0L,
							   (LPDWORD) 0) ) ) {

		sprintf(errMsg,"Debug thread failed to start");
		return 0;

	}
		
	  return 1;
}

int __stdcall AttachDebugger(int mpid){
	
   if(isDebugging==1){
		sprintf(errMsg,"Only one debuggee per session!");
		return 0;
   }
			
   pid=mpid;	
   proc[0]=0;
 
   if( !( hDebugEventThread = CreateThread(
 							   (LPSECURITY_ATTRIBUTES) 0L,
							   (DWORD) 0,
							   (LPTHREAD_START_ROUTINE) DebugEventThread,
							   (LPVOID) 0,
							   0L,
							   (LPDWORD) 0) ) ) {

		sprintf(errMsg,"Debug thread failed to start");
		return 0;

	}
		
	  return 1;
}


BOOL LaunchProcess(LPTSTR lpszFileName){

  static STARTUPINFO           StartupInfo;
  static LPSTARTUPINFO         lpStartupInfo = &StartupInfo;
  static PROCESS_INFORMATION   ProcessInfo;
  static LPPROCESS_INFORMATION lpProcessInfo = &ProcessInfo;
	
  memset(&StartupInfo,0,sizeof(STARTUPINFO));
  memset(&ProcessInfo,0,sizeof(PROCESS_INFORMATION));

  lpStartupInfo->cb          = sizeof( STARTUPINFO );
  lpStartupInfo->lpTitle     = "SandSprite VB Debugger Extension";
  lpStartupInfo->wShowWindow = SW_SHOWDEFAULT;

  lpProcessInfo->hProcess = 0L;
					  
  if( !CreateProcess(0L, lpszFileName,(LPSECURITY_ATTRIBUTES) 0L, 
					(LPSECURITY_ATTRIBUTES) 0L, TRUE, 
					 DEBUG_PROCESS | CREATE_NEW_CONSOLE,
					(LPVOID) 0L, (LPTSTR) 0L,lpStartupInfo,lpProcessInfo )
	 ) {

			switch( GetLastError() ) {
				  case ERROR_FILE_NOT_FOUND:
					sprintf(errMsg,"This file does not exist.");
					break;
				  case ERROR_ACCESS_DENIED:
					sprintf(errMsg,"Access denied.");
					break;
				  case ERROR_FILE_INVALID:
					sprintf(errMsg,"Invalid file." );
					break;
				  case ERROR_FILE_CORRUPT:
					sprintf(errMsg,"The file is corrupt." );
					break;
				  case ERROR_BAD_EXE_FORMAT:
					sprintf(errMsg,"The file has a bad format." );
					break;
				  default:
					break;
					//sprintf(errMsg,"CreateProcess()" );
			}
			return( FALSE );
  }
  else {
			pid = lpProcessInfo->dwProcessId; 
			//CloseHandle( ProcessInfo.hProcess );
			//CloseHandle( ProcessInfo.hThread );	
  }
  return(TRUE);
}





DWORD WINAPI DebugEventThread(void* x){

	

    if( pid==0 ){
		if( !LaunchProcess( proc ) ){
			sprintf(errMsg,"DebugNewProcess Failed...");
			isDebugging = 0;
			ExitThread( FALSE );
		}
	}
	else{
		  if( !DebugActiveProcess( pid ) ){
			 sprintf(errMsg,"DebugActiveProcess Failed pid %d", pid);
			 isDebugging = 0;
			 ExitThread( FALSE );
		  }
	}

    isDebugging = 1;

	for(;;) {

		if( !WaitForDebugEvent( &DebugEvent, INFINITE ) ) {
		  sprintf(errMsg,"Failed to attach to Debuggee...");
		  isDebugging = 0;
		  ExitThread(FALSE);
		  break;
		}
		
		debugDecision = 0;
		DbgMsg(&DebugEvent);		 
		
		while(debugDecision==0){
			Sleep(60);
		}

		ContinueDebugEvent( DebugEvent.dwProcessId,DebugEvent.dwThreadId, debugDecision );
	
		  
	}

	return 1;
}


// PURPOSE  : Retrieves the DLL module name for a given file handle of a
//            the module.  Reads the module name from the EXE header.
//
int __stdcall GetModuleFileNameFromHeader(int hProcess, HANDLE hFile, 
										  int BaseOfDll, char* lpszPath,
										  int bufSize)
{
  #define IMAGE_SECOND_HEADER_OFFSET    (15 * sizeof(ULONG)) // relative to file beginning
  #define IMAGE_BASE_OFFSET             (13 * sizeof(DWORD)) // relative to PE header base
  #define IMAGE_EXPORT_TABLE_RVA_OFFSET (30 * sizeof(DWORD)) // relative to PE header base
  #define IMAGE_NAME_RVA_OFFSET         offsetof(IMAGE_EXPORT_DIRECTORY, Name)
	

  WORD   DosSignature;
  DWORD  NtSignature;
  DWORD  dwNumberOfBytesRead = 0;
  DWORD  PeHeader;
  DWORD_PTR ImageBase, ExportTableRVA, NameRVA;

  //-- verify that the handle is not 0L
  if( !hFile ) {
    strcpy(lpszPath,"Invalid File Handle");
    return( 0 );
  }

  //-- verify that the handle is for a disk file
  if( GetFileType(hFile) != FILE_TYPE_DISK ) {
    strcpy(lpszPath,"Invalid File Type");
    return( 0 );
  }

  //-- Extract the filename from the EXE header
  SetFilePointer(hFile, 0L, 0L, FILE_BEGIN );
  ReadFile( hFile, &DosSignature, sizeof(DosSignature), &dwNumberOfBytesRead,(LPOVERLAPPED) 0L);

  //-- verify DOS signature found
  if( DosSignature != IMAGE_DOS_SIGNATURE ) {
    sprintf(lpszPath,"Bad MZ Signature: 0x%x", DosSignature );
    return( 0 );
  }

  SetFilePointer( hFile, IMAGE_SECOND_HEADER_OFFSET, (LPLONG) 0L,FILE_BEGIN );
  ReadFile( hFile, &PeHeader, sizeof(PeHeader), &dwNumberOfBytesRead,(LPOVERLAPPED) 0L );
  SetFilePointer( hFile, PeHeader, (LPLONG) 0L, FILE_BEGIN );
  ReadFile( hFile, &NtSignature, sizeof(NtSignature), &dwNumberOfBytesRead,(LPOVERLAPPED) 0L);

  //-- verify Windows NT (PE) signature found
  if( NtSignature != IMAGE_NT_SIGNATURE ) {
    sprintf(lpszPath,"Bad PE Signature: 0x%x",DosSignature);
    return( 0 );
  }

  SetFilePointer( hFile, PeHeader + IMAGE_BASE_OFFSET, (LPLONG) 0L,FILE_BEGIN );
  ReadFile( hFile, &ImageBase, sizeof(ImageBase), &dwNumberOfBytesRead, (LPOVERLAPPED) 0L);
  SetFilePointer( hFile, PeHeader + IMAGE_EXPORT_TABLE_RVA_OFFSET,(LPLONG) 0L, FILE_BEGIN );
  ReadFile( hFile, &ExportTableRVA, sizeof(ExportTableRVA), &dwNumberOfBytesRead, (LPOVERLAPPED) 0L);

  //-- now read from the virtual address space in the process
  if (!ReadProcessMemory( (HANDLE)hProcess,
         (LPVOID) (BaseOfDll + ExportTableRVA + IMAGE_NAME_RVA_OFFSET),
         &NameRVA, sizeof(NameRVA), &dwNumberOfBytesRead ) ||
      dwNumberOfBytesRead != sizeof(NameRVA))
  {
      strcpy(lpszPath,"Access Denied!");
      dwNumberOfBytesRead = 0;
  }
  else if( !ReadProcessMemory( (HANDLE) hProcess,
              (LPVOID) (BaseOfDll + NameRVA), lpszPath, bufSize, &dwNumberOfBytesRead ))
  {
     strcpy(lpszPath,"Access Denied!");
  }

  return( dwNumberOfBytesRead );
}

